@@ -1,168 +1,179 @@
-use std::{collections::HashSet, convert::TryFrom, path::Path};
+use std::{collections::HashSet, convert::TryFrom, path::Path, path::PathBuf};
 
 use html5ever::{
     parse_document,
     tendril::{StrTendril, TendrilSink},
 };
-use linkify::LinkFinder;
+use log::info;
 use markup5ever_rcdom::{Handle, NodeData, RcDom};
+use percent_encoding::percent_decode_str;
 use pulldown_cmark::{Event as MDEvent, Parser, Tag};
-use url::Url;
+use reqwest::Url;
 
-use crate::{collector::InputContent, Request, Uri};
+use crate::{
+    helpers::{path, url},
+    types::{FileType, InputContent},
+    Base, ErrorKind, Input, Request, Result, Uri,
+};
 
-#[derive(Copy, Clone, Debug, PartialEq, Eq)]
-/// `FileType` defines which file types lychee can handle
-pub enum FileType {
-    /// File in HTML format
-    Html,
-    /// File in Markdown format
-    Markdown,
-    /// Generic text file without syntax-specific parsing
-    Plaintext,
+/// A handler for extracting links from various input formats like Markdown and
+/// HTML. Allocations are avoided if possible as this is a performance-critical
+/// section of the library.
+#[derive(Debug)]
+pub struct Extractor {
+    /// URLs extracted from input
+    pub urls: Vec<StrTendril>,
+    /// Base URL or Path
+    pub base: Option<Base>,
 }
 
-impl Default for FileType {
-    fn default() -> Self {
-        Self::Plaintext
-    }
-}
-
-impl<P: AsRef<Path>> From<P> for FileType {
-    /// Detect if the given path points to a Markdown, HTML, or plaintext file.
-    fn from(p: P) -> FileType {
-        let path = p.as_ref();
-        // Assume HTML in case of no extension.
-        // Note: this is only reasonable for URLs; not paths on disk.
-        // For example, `README` without an extension is more likely to be a plaintext file.
-        // A better solution would be to also implement `From<Url> for FileType`.
-        // Unfortunately that's not possible without refactoring, as
-        // `AsRef<Path>` could be implemented for `Url` in the future, which is why
-        // `From<Url> for FileType` is not allowed.
-        match path.extension().and_then(std::ffi::OsStr::to_str) {
-            Some("md" | "markdown") => FileType::Markdown,
-            Some("htm" | "html") | None => FileType::Html,
-            Some(_) => FileType::Plaintext,
+impl Extractor {
+    pub(crate) const fn new(base: Option<Base>) -> Self {
+        Extractor {
+            urls: Vec::new(),
+            base,
         }
     }
-}
 
-// Use LinkFinder here to offload the actual link searching in plaintext.
-fn find_links(input: &str) -> Vec<linkify::Link> {
-    let finder = LinkFinder::new();
-    finder.links(input).collect()
-}
+    /// Main entrypoint for extracting links from various sources
+    /// (Markdown, HTML, and plaintext)
+    pub(crate) fn extract(&mut self, input_content: &InputContent) -> Result<HashSet<Request>> {
+        match input_content.file_type {
+            FileType::Markdown => self.extract_markdown(&input_content.content),
+            FileType::Html => self.extract_html(&input_content.content),
+            FileType::Plaintext => self.extract_plaintext(&input_content.content),
+        };
+        self.create_requests(input_content)
+    }
 
-/// Extract unparsed URL strings from a markdown string.
-fn extract_links_from_markdown(input: &str) -> Vec<String> {
-    let parser = Parser::new(input);
-    parser
-        .flat_map(|event| match event {
-            MDEvent::Start(Tag::Link(_, url, _) | Tag::Image(_, url, _)) => {
-                vec![url.to_string()]
+    /// Create requests out of the collected URLs.
+    /// Only keeps legit URLs. For example this filters out anchors.
+    fn create_requests(&self, input_content: &InputContent) -> Result<HashSet<Request>> {
+        let mut requests: HashSet<Request> = HashSet::with_capacity(self.urls.len());
+
+        let base_input = match &input_content.input {
+            Input::RemoteUrl(url) => Some(Url::parse(&format!(
+                "{}://{}",
+                url.scheme(),
+                url.host_str().ok_or(ErrorKind::InvalidUrlHost)?
+            ))?),
+            _ => None,
+            // other inputs do not have a URL to extract a base
+        };
+
+        for url in &self.urls {
+            let req = if let Ok(uri) = Uri::try_from(url.as_ref()) {
+                Request::new(uri, input_content.input.clone())
+            } else if let Some(url) = self.base.as_ref().and_then(|u| u.join(url)) {
+                Request::new(Uri { url }, input_content.input.clone())
+            } else if let Input::FsPath(root) = &input_content.input {
+                if url::is_anchor(url) {
+                    // Silently ignore anchor links for now
+                    continue;
+                }
+                match self.create_uri_from_path(root, url)? {
+                    Some(url) => Request::new(Uri { url }, input_content.input.clone()),
+                    None => {
+                        // In case we cannot create a URI from a path but we didn't receive an error,
+                        // it means that some preconditions were not met, e.g. the `base_url` wasn't set.
+                        continue;
+                    }
+                }
+            } else if let Some(url) = base_input.as_ref().map(|u| u.join(url)) {
+                if self.base.is_some() {
+                    continue;
+                }
+                Request::new(Uri { url: url? }, input_content.input.clone())
+            } else {
+                info!("Handling of {} not implemented yet", &url);
+                continue;
+            };
+            requests.insert(req);
+        }
+        Ok(requests)
+    }
+
+    /// Extract unparsed URL strings from a Markdown string.
+    fn extract_markdown(&mut self, input: &str) {
+        let parser = Parser::new(input);
+        for event in parser {
+            match event {
+                MDEvent::Start(Tag::Link(_, url, _) | Tag::Image(_, url, _)) => {
+                    self.urls.push(StrTendril::from(url.as_ref()));
+                }
+                MDEvent::Text(txt) => self.extract_plaintext(&txt),
+                MDEvent::Html(html) => self.extract_html(&html.to_string()),
+                _ => {}
             }
-            MDEvent::Text(txt) => extract_links_from_plaintext(&txt.to_string()),
-            MDEvent::Html(html) => extract_links_from_html(&html.to_string()),
-            _ => vec![],
-        })
-        .collect()
-}
-
-/// Extract unparsed URL strings from a HTML string.
-fn extract_links_from_html(input: &str) -> Vec<String> {
-    let tendril = StrTendril::from(input);
-    let rc_dom = parse_document(RcDom::default(), html5ever::ParseOpts::default()).one(tendril);
-
-    let mut urls = Vec::new();
-
-    // we pass mutable urls reference to avoid extra allocations in each
-    // recursive descent
-    walk_html_links(&mut urls, &rc_dom.document);
-
-    urls
-}
-
-/// Recursively walk links in a HTML document, aggregating URL strings in `urls`.
-fn walk_html_links(mut urls: &mut Vec<String>, node: &Handle) {
-    match node.data {
-        NodeData::Text { ref contents } => {
-            urls.append(&mut extract_links_from_plaintext(&contents.borrow()));
         }
+    }
 
-        NodeData::Comment { ref contents } => {
-            urls.append(&mut extract_links_from_plaintext(contents));
-        }
+    /// Extract unparsed URL strings from an HTML string.
+    fn extract_html(&mut self, input: &str) {
+        let tendril = StrTendril::from(input);
+        let rc_dom = parse_document(RcDom::default(), html5ever::ParseOpts::default()).one(tendril);
 
-        NodeData::Element {
-            ref name,
-            ref attrs,
-            ..
-        } => {
-            for attr in attrs.borrow().iter() {
-                let attr_value = attr.value.to_string();
+        self.walk_html_links(&rc_dom.document);
+    }
 
-                if elem_attr_is_link(attr.name.local.as_ref(), name.local.as_ref()) {
-                    urls.push(attr_value);
-                } else {
-                    urls.append(&mut extract_links_from_plaintext(&attr_value));
+    /// Recursively walk links in a HTML document, aggregating URL strings in `urls`.
+    fn walk_html_links(&mut self, node: &Handle) {
+        match node.data {
+            NodeData::Text { ref contents } => {
+                self.extract_plaintext(&contents.borrow());
+            }
+            NodeData::Comment { ref contents } => {
+                self.extract_plaintext(contents);
+            }
+            NodeData::Element {
+                ref name,
+                ref attrs,
+                ..
+            } => {
+                for attr in attrs.borrow().iter() {
+                    if url::elem_attr_is_link(attr.name.local.as_ref(), name.local.as_ref()) {
+                        self.urls.push(attr.value.clone());
+                    } else {
+                        self.extract_plaintext(&attr.value);
+                    }
                 }
             }
+            _ => {}
         }
 
-        _ => {}
+        // recursively traverse the document's nodes -- this doesn't need any extra
+        // exit conditions, because the document is a tree
+        for child in node.children.borrow().iter() {
+            self.walk_html_links(child);
+        }
     }
 
-    // recursively traverse the document's nodes -- this doesn't need any extra
-    // exit conditions because the document is a tree
-    for child in node.children.borrow().iter() {
-        walk_html_links(&mut urls, child);
+    /// Extract unparsed URL strings from plaintext
+    fn extract_plaintext(&mut self, input: &str) {
+        self.urls
+            .extend(url::find_links(input).map(|l| StrTendril::from(l.as_str())));
     }
-}
 
-/// Determine if element's attribute contains a link / URL.
-fn elem_attr_is_link(attr_name: &str, elem_name: &str) -> bool {
-    // See a comprehensive list of attributes that might contain URLs/URIs
-    // over at: https://developer.mozilla.org/en-US/docs/Web/HTML/Attributes
-    matches!(
-        (attr_name, elem_name),
-        ("href" | "src" | "srcset" | "cite", _) | ("data", "object") | ("onhashchange", "body")
-    )
-}
-
-/// Extract unparsed URL strings from a plaintext.
-fn extract_links_from_plaintext(input: &str) -> Vec<String> {
-    find_links(input)
-        .iter()
-        .map(|l| String::from(l.as_str()))
-        .collect()
-}
-
-pub(crate) fn extract_links(
-    input_content: &InputContent,
-    base_url: &Option<Url>,
-) -> HashSet<Request> {
-    let links = match input_content.file_type {
-        FileType::Markdown => extract_links_from_markdown(&input_content.content),
-        FileType::Html => extract_links_from_html(&input_content.content),
-        FileType::Plaintext => extract_links_from_plaintext(&input_content.content),
-    };
-
-    // Only keep legit URLs. This sorts out things like anchors.
-    // Silently ignore the parse failures for now.
-    let mut requests: HashSet<Request> = HashSet::new();
-    for link in links {
-        if let Ok(uri) = Uri::try_from(link.as_str()) {
-            requests.insert(Request::new(uri, input_content.input.clone()));
-        } else if !Path::new(&link).exists() {
-            if let Some(new_url) = base_url.as_ref().and_then(|u| u.join(&link).ok()) {
-                requests.insert(Request::new(
-                    Uri { url: new_url },
-                    input_content.input.clone(),
-                ));
-            }
-        };
+    fn create_uri_from_path(&self, src: &Path, dst: &str) -> Result<Option<Url>> {
+        let dst = url::remove_get_params_and_fragment(dst);
+        // Avoid double-encoding already encoded destination paths by removing any
+        // potential encoding (e.g. `web%20site` becomes `web site`).
+        // That's because Url::from_file_path will encode the full URL in the end.
+        // This behavior cannot be configured.
+        // See https://github.com/lycheeverse/lychee/pull/262#issuecomment-915245411
+        // TODO: This is not a perfect solution.
+        // Ideally, only `src` and `base` should be URL encoded (as is done by
+        // `from_file_path` at the moment) while `dst` is left untouched and simply
+        // appended to the end.
+        let decoded = percent_decode_str(dst).decode_utf8()?;
+        let resolved = path::resolve(src, &PathBuf::from(&*decoded), &self.base)?;
+        match resolved {
+            Some(path) => Url::from_file_path(&path)
+                .map(Some)
+                .map_err(|_e| ErrorKind::InvalidUrlFromPath(path)),
+            None => Ok(None),
+        }
     }
-    requests
 }
 
 #[cfg(test)]
@@ -176,17 +187,26 @@ mod test {
     };
 
     use pretty_assertions::assert_eq;
-    use url::Url;
 
-    use super::{
-        extract_links, extract_links_from_html, extract_links_from_markdown,
-        extract_links_from_plaintext, find_links, FileType,
-    };
+    use super::*;
     use crate::{
-        collector::InputContent,
+        helpers::url::find_links,
         test_utils::{mail, website},
         Uri,
     };
+    use crate::{
+        types::{FileType, InputContent},
+        Base,
+    };
+
+    #[test]
+    fn test_create_uri_from_path() {
+        let extractor = Extractor::new(None);
+        let result = extractor
+            .create_uri_from_path(&PathBuf::from("/README.md"), "test+encoding")
+            .unwrap();
+        assert_eq!(result.unwrap().as_str(), "file:///test+encoding");
+    }
 
     fn load_fixture(filename: &str) -> String {
         let fixture_path = Path::new(env!("CARGO_MANIFEST_DIR"))
@@ -207,13 +227,15 @@ mod test {
     }
 
     fn extract_uris(input: &str, file_type: FileType, base_url: Option<&str>) -> HashSet<Uri> {
-        extract_links(
-            &InputContent::from_string(input, file_type),
-            &base_url.map(|u| Url::parse(u).unwrap()),
-        )
-        .into_iter()
-        .map(|r| r.uri)
-        .collect()
+        let base = base_url.map(|url| Base::Remote(Url::parse(url).unwrap()));
+        let mut extractor = Extractor::new(base);
+        extractor
+            .extract(&InputContent::from_string(input, file_type))
+            // unwrap is fine here as this helper function is only used in tests
+            .unwrap()
+            .into_iter()
+            .map(|r| r.uri)
+            .collect()
     }
 
     #[test]
@@ -233,12 +255,20 @@ mod test {
 
     #[test]
     fn test_extract_link_at_end_of_line() {
-        let input = "http://www.apache.org/licenses/LICENSE-2.0\n";
+        let input = "https://www.apache.org/licenses/LICENSE-2.0\n";
         let link = input.trim_end();
 
-        assert_eq!(vec![link], extract_links_from_markdown(input));
-        assert_eq!(vec![link], extract_links_from_plaintext(input));
-        assert_eq!(vec![link], extract_links_from_html(input));
+        let mut extractor = Extractor::new(None);
+        extractor.extract_markdown(input);
+        assert_eq!(vec![StrTendril::from(link)], extractor.urls);
+
+        let mut extractor = Extractor::new(None);
+        extractor.extract_plaintext(input);
+        assert_eq!(vec![StrTendril::from(link)], extractor.urls);
+
+        let mut extractor = Extractor::new(None);
+        extractor.extract_html(input);
+        assert_eq!(vec![StrTendril::from(link)], extractor.urls);
     }
 
     #[test]
@@ -343,7 +373,7 @@ mod test {
     #[test]
     fn test_md_escape() {
         let input = r#"http://msdn.microsoft.com/library/ie/ms535874\(v=vs.85\).aspx"#;
-        let links = find_links(input);
+        let links: Vec<_> = find_links(input).collect();
         let expected = "http://msdn.microsoft.com/library/ie/ms535874(v=vs.85).aspx)";
 
         matches!(&links[..], [link] if link.as_str() == expected);
@@ -386,6 +416,42 @@ mod test {
         assert_eq!(links, expected_links);
     }
 
+    #[test]
+    fn test_relative_url_with_base_extracted_from_input() {
+        let input = Input::RemoteUrl(Box::new(
+            Url::parse("https://example.org/some-post").unwrap(),
+        ));
+
+        let contents = r#"<html>
+            <div class="row">
+                <a href="https://github.com/lycheeverse/lychee/">Github</a>
+                <a href="/about">About</a>
+            </div>
+        </html>"#;
+
+        let input_content = &InputContent {
+            input,
+            file_type: FileType::Html,
+            content: contents.to_string(),
+        };
+
+        let mut extractor = Extractor::new(None);
+        let links = extractor.extract(input_content);
+        let urls = links
+            .unwrap()
+            .iter()
+            .map(|x| x.uri.url.as_str().to_string())
+            .collect::<HashSet<_>>();
+
+        let expected_urls = array::IntoIter::new([
+            String::from("https://github.com/lycheeverse/lychee/"),
+            String::from("https://example.org/about"),
+        ])
+        .collect::<HashSet<_>>();
+
+        assert_eq!(urls, expected_urls);
+    }
+
     #[test]
     fn test_extract_html5_lowercase_doctype() {
         // this has been problematic with previous XML based parser
